/*
 * ----------------------------------------------------------------------------
 *  (C) Copyright Testing Technologies, 2001-2005.  All Rights Reserved.
 *
 *  All copies of this program, whether in whole or in part, and whether
 *  modified or not, must display this and all other embedded copyright
 *  and ownership notices in full.
 *
 *  See the file COPYRIGHT for details of redistribution and use.
 *
 *  You should have received a copy of the COPYRIGHT file along with
 *  this file; if not, write to the Testing Technologies,
 *  Rosenthaler Str. 13, 10119 Berlin, Germany.
 *
 *  TESTING TECHNOLOGIES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *  SOFTWARE. IN NO EVENT SHALL TESTING TECHNOLOGIES BE LIABLE FOR ANY
 *  SPECIAL, DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 *  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 *  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *  THIS SOFTWARE.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 *  EITHER EXPRESSED OR IMPLIED, INCLUDING ANY KIND OF IMPLIED OR
 *  EXPRESSED WARRANTY OF NON-INFRINGEMENT OR THE IMPLIED WARRANTIES
 *  OF MECHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * -----------------------------------------------------------------------------
 *
 *  AUTHOR:      ldte
 *  DATE:        Mar 15, 2005
 *
 *  REVISION INFO:
 *    $Revision: 1.5 $ $Date: 2005/09/13 08:47:39 $
 *    $Source: /net/ttech/cvs_root/IPv6_ETSI_STF276/javasrc/com/testingtech/ttcn/tri/JpcapRunnable.java,v $
 *
 * -----------------------------------------------------------------------------
 */


package com.testingtech.ttcn.tri;

import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import jpcap.Jpcap;
import jpcap.Packet;

import org.etsi.ttcn.tri.TriComponentId;
import org.etsi.ttcn.tri.TriMessage;
import org.etsi.ttcn.tri.TriPortId;

import com.testingtech.util.HexViewer;

import de.tu_berlin.cs.uebb.muttcn.runtime.RB;

/**
 * The intention of this class is to handle all the received messages 
 * and enqueue them based on the information of the MAC address and 
 * the rcvMap to the TTCN-3 components.
 * In case of well known addresses the specific handling(e.g. enqueuing to 
 * all components) is done.
 * 
 * @author ldte
 *
 */
public class JpcapRunnable implements Runnable, TestAdapterDefinitions {
	
	private Jpcap f_jpcap = null;
	private String f_netDeviceId = "";
	
	private HashMap f_portMap = null;
	private HashMap f_rcvHashMap = null;
	
	private boolean runThread = false;
//	private boolean isRunning = false;
	
	private RB RB = null;
	
	/**
	 * The constructor of the class.
	 * 
	 * @param rb			The runtime used.
	 * @param jpcap			The Jpcap instance for capturing.
	 * @param netDeviceId	The net device used.
	 * @param portMap		The portMap including the component and tsiPort information.
	 * @param rcvMap		The receive hashmap inlcuding all information for the enqueuing(together with th portMap).
	 */
	public JpcapRunnable(RB rb, Jpcap jpcap, String netDeviceId, HashMap portMap, HashMap rcvMap) {
		super();
		RB = rb;
		f_jpcap = jpcap;
		f_netDeviceId = netDeviceId;
		f_portMap = portMap;
		f_rcvHashMap = rcvMap;
	}
	
	/**
	 * @param map The f_portMap to set.
	 */
	public synchronized void setPortMap(HashMap map) {
		f_portMap = map;
	}

	/**
	 * @param hashMap The f_rcvHashMap to set.
	 */
	public synchronized void setRcvHashMap(HashMap hashMap) {
		f_rcvHashMap = hashMap;
	}

	private synchronized boolean getRunThread() {
		return runThread;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		synchronized (this) {
			runThread = true;
			notifyAll();
//			isRunning = true;
		}

		while (getRunThread()) {
			try {
				Thread.sleep(100);
				Packet packet = f_jpcap.getPacket();

				if (packet == null) {
					continue;
				}
				
				if (getRunThread()) {
					byte[] rcvmsg = packet.data;
					TriMessage trmsg = new TriMessageImpl(rcvmsg);
					
					if (RB.debug) {
						System.out.println(HexViewer.hexString(packet.header));
						System.out.println(HexViewer.hexString(packet.data));
					}
					
					byte [] ipv6SrcAddress = new byte[16];
					byte [] ipv6DstAddress = new byte[16];
					byte [] ipv6SolNodeAddress = new byte[13];
					byte [] macSrcAddress = new byte[6];
					byte [] macDstAddress = new byte[6];
					
					System.arraycopy(packet.data,  8, ipv6SrcAddress, 0, 16);
					System.arraycopy(packet.data, 24, ipv6DstAddress, 0, 16);
					System.arraycopy(packet.data, 24, ipv6SolNodeAddress, 0, 13);
					System.arraycopy(packet.header, 0, macDstAddress, 0, 6);
					System.arraycopy(packet.header, 6, macSrcAddress, 0, 6);
					
					/*
					 * If the received IPv6 message is a broadcast message, enqueue the message to every known component
					 */
					if (Arrays.equals(ipv6DstAddress, IP_ALLNODESMCA)) { 
//						if (AllMcaReceivedFromSUT(packet.data) == true) {
							synchronized (f_portMap) {
								for (Iterator it=f_portMap.entrySet().iterator(); it.hasNext(); ) {
									Map.Entry entry = (Map.Entry)it.next();
									TriPortId compPortId = ((TriPortId)entry.getKey());
									
									TriPortId tsiPortId = (TriPortId)f_portMap.get(compPortId);
									
									//    	                    		System.out.println("Enqueuing IP_ALLNODESMCA to "+ compPortId.getComponent().getComponentId());
									
									if (tsiPortId != null) {
										((IPv6TestAdapter)RB.TestAdapter).enqueueMsg(tsiPortId, new TriAddressImpl(new byte[] {  }),
												compPortId.getComponent(), trmsg);
									}
									else {
										((IPv6TestAdapter)RB.TestAdapter).log(TSIPORTID_NOT_FOUND + compPortId);
										continue;
									}
								}
							}
//						}
					}
					/*
					 * If the received IPv6 message is router broadcast message, enqueue the message to every known router component
					 */
					else if (Arrays.equals(ipv6DstAddress, IP_ALLROUTERSMCA) || 
							(Arrays.equals(ipv6SolNodeAddress, IP_SOLNODEMCA) && Arrays.equals(ipv6SrcAddress, IP_UNSPECIFIED))
							) {
//						if (AllMcaReceivedFromSUT(packet.data) == true) {
							synchronized (f_rcvHashMap) {
								synchronized (f_portMap) {
									for (Iterator it=f_portMap.entrySet().iterator(); it.hasNext(); ) {
										Map.Entry entry = (Map.Entry)it.next();
										TriComponentId componentId = ((TriPortId)entry.getKey()).getComponent();
										
										for (Iterator rit=f_rcvHashMap.entrySet().iterator(); rit.hasNext(); ) {
											Map.Entry rentry = (Map.Entry)rit.next();
											PTCAndRole par = ((PTCAndRole)rentry.getValue());
											
											if ((par != null) && (par.getComponentId().equals(componentId)) && (par.getRole().startsWith(ROUTER_ROLE))) {
												TriPortId compPortId = new TriPortIdImpl(IPPORT, componentId, -1);
												
												TriPortId tsiPortId = (TriPortId)f_portMap.get(compPortId);
												
												//		                    		System.out.println("Enqueuing IP_ALLROUTERSMCA to "+ compPortId.getComponent().getComponentId());
												
												if (tsiPortId != null) {
													((IPv6TestAdapter)RB.TestAdapter).enqueueMsg(tsiPortId, new TriAddressImpl(new byte[] {  }),
															componentId, trmsg);
												}
												else {
													((IPv6TestAdapter)RB.TestAdapter).log(TSIPORTID_NOT_FOUND + compPortId);
													continue;
												}
												//we have found the component with role router, so don't look for anymore
												break;
											}
										}
									}
								}
							}
//						}
						//                        for (Iterator it=f_rcvHashMap.entrySet().iterator(); it.hasNext(); ) {
						//                            Map.Entry entry = (Map.Entry)it.next();
						//                            PTCAndRole par = ((PTCAndRole)entry.getValue());
						//
						//                            if ((par != null) && (par.getRole().startsWith(ROUTER_ROLE))) {
						//                            	TriPortId compPortId = new TriPortIdImpl(IPPORT, par.getComponentId(), -1);
						//                            	
						//                            	TriPortId tsiPortId = (TriPortId)f_portMap.get(compPortId);
						//                            	
						////	                    		System.out.println("Enqueuing IP_ALLROUTERSMCA to "+ compPortId.getComponent().getComponentId());
						//	                    		
						//                            	if (tsiPortId != null) {
						//                            		((IPv6TestAdapter)RB.TestAdapter).enqueueMsg(tsiPortId, new TriAddressImpl(new byte[] {  }),
						//                            				par.getComponentId(), trmsg);
						//                            	}
						//                            	else {
						//                            		((IPv6TestAdapter)RB.TestAdapter).log(TSIPORTID_NOT_FOUND + compPortId);
						//                            		continue;
						//                            	}
						//                            }
						//                    	}
					}
					/*
					 * Enqueue the received IPv6 message, enqueue the message to the configured component based on the MAC address
					 */
					else {
//						if (AllMcaReceivedFromSUT(packet.data) == true) {
//						if (AllMcaReceivedFromSUT(macSrcAddress) == true) {
							synchronized (f_rcvHashMap) {
								PTCAndRole par = (PTCAndRole)f_rcvHashMap.get(new BytesHash(macDstAddress));
								if (par == null) {
									((IPv6TestAdapter)RB.TestAdapter).log(TRIENQUEUEMSG_MACADDRESS_NOT_CONFIGURED + HexViewer.hexString(macDstAddress));
								}
								else {
									TriComponentId componentId = par.getComponentId();
									if (componentId != null) {
										TriPortId compPortId = new TriPortIdImpl(IPPORT, componentId, -1);
										
										TriPortId tsiPortId = null;
										synchronized (f_portMap) {
											tsiPortId = (TriPortId)f_portMap.get(compPortId);
										}
										
										if (RB.debug) {
											System.out.println("Enqueuing to "+ compPortId.getComponent().getComponentId());
											System.out.println(HexViewer.hexString(rcvmsg));
										}
										
										if (tsiPortId != null) {
											((IPv6TestAdapter)RB.TestAdapter).enqueueMsg(tsiPortId, new TriAddressImpl(new byte[] {  }),
													componentId, trmsg);
										}
										else {
											((IPv6TestAdapter)RB.TestAdapter).log(TSIPORTID_NOT_FOUND + compPortId);
											continue;
										}
									}
									else {
										((IPv6TestAdapter)RB.TestAdapter).log(TRIENQUEUEMSG_NO_COMPONENT_FOUND 
												+ ((IPv6TestAdapter)RB.TestAdapter).getIPv6AddressString(ipv6DstAddress) 
												+ "(" + f_netDeviceId + ")");
									}
								}
							}
//						}
					}
				}
			}
			catch (InterruptedException iex) {}
		}
	    //TODO
		SimpleDateFormat df = new SimpleDateFormat("HH:mm:ss:SSS");
		Calendar cal = Calendar.getInstance();
    	cal.setTimeInMillis(System.currentTimeMillis());
    	if (RB.debug)
    		System.err.println(df.format(cal.getTime()) + "----------------------------------------thread end-----------------------------------" + this);
//		System.out.println("Closing jpcap");
//		f_jpcap.myClose();
		synchronized (this) {
//			isRunning = false;
			notifyAll();
		}
	}
	
	/**
	 * Returns current running state.
	 *
	 * @return true if running, false otherwise
	 */
//	public synchronized boolean isRunning() {
//		return isRunning;
//	}
	
	/**
	 * Cancels the running receiver thread
	 */
	public synchronized void cancel() {
		while (!runThread) {
			try {
				wait();
			} catch (InterruptedException e) {
			}
		}
		synchronized (this) {
			runThread = false;
		}
	}
	
	/**
	 * Checks if the received message is not the one we sent before, based on the MAC adresses.
	 * 
	 * To avoid that messages with multicast address sent bt PTC are enqueued to the PTC directly,
	 * all received messages on the netword card are checked that their src-address is not the PTC src-address
	 * 
	 * @param packetdata	The message to check
	 * @return				True in case of the message is not the one sent, false otherwise
	 */
//changed to use mac address instead of ipv6 one
	private synchronized boolean AllMcaReceivedFromSUT(byte [] packetdata) {
////		byte [] srcAddressLastTwoOctets = new byte[2];
////		System.arraycopy(packetdata, 22, srcAddressLastTwoOctets, 0, 2);
//		
//		OctetstringValue ov = null;
//		
////		for (int i=0; i<PX_LLA_X.length; i++) {
////			ov = (OctetstringValue)RB.tciTMProvided.tciGetModulePar(TciModuleParameterIdImpl.valueOf(PX_LLA_X[i]));
////			if (ov == null) {
////				RB.tciTMProvided.tciError(MODULEPAR_NOT_FOUND + PX_LLA_X[i]);
////				System.err.println(MODULEPAR_NOT_FOUND + PX_LLA_X[i]);
////				return false;
////			}
//		for (int i=0; i<PX_MAC_UCA_X.length; i++) {
//			ov = (OctetstringValue)RB.tciTMProvided.tciGetModulePar(TciModuleParameterIdImpl.valueOf(PX_MAC_UCA_X[i]));
//			if (ov == null) {
//				RB.tciTMProvided.tciError(MODULEPAR_NOT_FOUND + PX_MAC_UCA_X[i]);
//				System.err.println(MODULEPAR_NOT_FOUND + PX_MAC_UCA_X[i]);
//				return false;
//			}
//			else {
//				OctetString o = new OctetString(RB, ov.getString());
//				byte [] ba = o.toByteArray();
////				byte [] ba = { (byte)ov.getOctet(ov.getLength()-2), (byte)ov.getOctet(ov.getLength()-1)};
////				if (Arrays.equals(srcAddressLastTwoOctets, ba)) {
//				if (Arrays.equals(packetdata, ba)) {
//					//        			System.err.println("@@@@@@@@@@@@@@@@@    " + HexViewer.hexString(srcAddressLastTwoOctets) + " @@@@@@@@@@@@@@@@@@@@@ " + HexViewer.hexString(ba));
//					return false;
//				}
//			}
//		}
//		
//		return true;
		if (RB.debug) {
			System.err.println("packetdata: " + DebugHelp.hexString(packetdata));
		}
		for (Iterator rit=f_rcvHashMap.keySet().iterator(); rit.hasNext(); ) {
			BytesHash ba = (BytesHash)rit.next();
			System.err.println("rcvHashMap: " + DebugHelp.hexString(ba.getBytes()));
			if (Arrays.equals(packetdata, ba.getBytes())) {
				return false;
			}
		}
		
		return true;
	}
}
